import pandas as pd
from wllib2.db_lib import db_utils
import io
from PIL import Image
from PyPDF2 import PdfFileReader
import base64
from azure.identity import ClientSecretCredential
from azure.keyvault.secrets import SecretClient


class AzTestHandler():

    def __init__(self, app_env, input_args, input_file=None):
        self.env = app_env
        self.args = input_args
        self.file = input_file
        return 

    def get_env_name(self):
        return f"This function is executing on the  {self.env.get_env_name()}!"

    def say_hello(self):
        name = self.args.get('Name')
        return f"Simon's Azure function says hello {name}!"

    def maths_numbers(self):
        a = self.args.get('a')
        b = self.args.get('b')
        sum_response = {}
        sum_response['Sum'] = int(a)+int(b)
        sum_response['Minus'] = int(a)-int(b)
        sum_response['Product'] = int(a)*int(b)
        sum_response['Divide'] = int(a)/int(b)
        return sum_response

    def get_simple_dataframe(self):
        raw_data = {'first_name': ['Jason', 'Molly', 'Tina', 'Jake', 'Amy'], 
                'last_name': ['Miller', 'Jacobson', 'Ali', 'Milner', 'Cooze'], 
                'age': [42, 52, 36, 24, 73], 
                'preTestScore': [4, 24, 31, 2, 3],
                'postTestScore': [25, 94, 57, 62, 70]}
        df = pd.DataFrame(raw_data, columns = ['first_name', 'last_name', 'age', 'preTestScore', 'postTestScore'])
        return df

    def analyse_pdf(self):
        response_dict = {}
        pdf = PdfFileReader(io.BytesIO(self.file))
        information = pdf.getDocumentInfo()
        number_of_pages = pdf.getNumPages()
        response_dict['information'] = information
        response_dict['number_of_pages'] = number_of_pages
        return response_dict


    def image_converter(self):
        data = {}

        img64 = self.args.get('InputImg')
        new_img_format = self.args.get('NewImageFormat')
        new_img_name = self.args.get('NewImageName')

        imgbytes = base64.b64decode(img64)
        img = Image.open(io.BytesIO(imgbytes))

        imgbytes_new = image_to_byte_array(img,new_img_format)

        img64_new = base64.b64encode(imgbytes_new).decode("utf-8")

        data['fil'] = img64_new
        data['fnam'] = new_img_name
        data['height'] = img.height
        data['width'] = img.width
        return data

    def get_test_keyvault_value(self):
        response_dict = {}
        secret_key_name = self.args.get('SecretKeyName')
        key_val = self.env.get_secret_value(secret_key_name)
        response_dict[secret_key_name] = key_val
        return response_dict

    def test_dbquery(self):
        db = self.args.get('DB')
        sqlquery = self.args.get('QRY')
        sqldb = db_utils.SQLDBConn(self.env, db_name=db)
        if not sqlquery:
            sqlquery = f'select * from Test.TestTable'
        df = sqldb.fetch_sql_query(sqlquery)
        return df


def image_to_byte_array(image:Image, img_format):
  imgByteArr = io.BytesIO()
  image.save(imgByteArr, format=img_format)
  imgByteArr = imgByteArr.getvalue()
  return imgByteArr



import io
from PIL import Image, ImageOps
import PyPDF2
from PyPDF2 import PdfFileReader
import base64
from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
from wllib2.pdf_lib import pdf_utils


def create_pdf_func_response(pdf_file, cog_serv_endpoint, cog_serv_key, max_pages):
    resp_dict = {}
    pdf_text, n_pages, n_pages_in_pdf = pdf_utils.get_text_from_pdfbytes_images(pdf_file, cog_serv_endpoint, cog_serv_key, max_pages = max_pages)
    resp_dict['PDFText'] = pdf_text
    resp_dict['NumberOfPDFPagesScanned'] = n_pages
    resp_dict['TotalPagesInPDF'] = n_pages_in_pdf
    return resp_dict

class PDFAzHandler():

    def __init__(self, app_env, input_args, input_file=None):
        self.env = app_env
        self.args = input_args
        self.file = input_file
        return 

    def get_binary_pdf_info(self):
        response_dict = {}
        pdf = PdfFileReader(io.BytesIO(self.file))
        information = pdf.getDocumentInfo()
        number_of_pages = pdf.getNumPages()
        response_dict['information'] = information
        response_dict['number_of_pages'] = number_of_pages
        return response_dict

    def get_json_pdf_info(self):
        response_dict = {}
        pdf_file_base64 = self.args.get('PDFFile')
        pdf_file_name = self.args.get('PDFFileName')
        pdf_file = base64.b64decode(pdf_file_base64)
        pdf = PdfFileReader(io.BytesIO(pdf_file))
        information = pdf.getDocumentInfo()
        number_of_pages = pdf.getNumPages()
        response_dict['FileName'] = pdf_file_name
        response_dict['information'] = information
        response_dict['number_of_pages'] = number_of_pages
        return response_dict

    def get_blob_pdf_info(self):
        response_dict = {}
        connect_str = self.env.get_parameter('AzStorageContainer')
        local_file_name = self.args.get('PDFFile')
        container_name = self.args.get('StorageSubDir')
        blob_service_client = BlobServiceClient.from_connection_string(connect_str)
        blob_client = blob_service_client.get_blob_client(container=container_name, blob=local_file_name)
        fstream = blob_client.download_blob()
        pdf_file = fstream.readall()
        pdf = PdfFileReader(io.BytesIO(pdf_file))
        information = pdf.getDocumentInfo()
        number_of_pages = pdf.getNumPages()
        response_dict['PDF'] = local_file_name
        response_dict['Blob'] = container_name
        response_dict['information'] = information
        response_dict['number_of_pages'] = number_of_pages
        return response_dict

    def get_pdffile_images_text(self):
        max_pages = -1
        max_pages_str = self.args.get('MaxPages')
        if max_pages_str is not None:
            max_pages = int(max_pages_str)
        cog_serv_endpoint = self.env.get_secret_value('ImgCogServicesEndpoint')
        cog_serv_key = self.env.get_secret_value('ImgCogServicesKey')
        return create_pdf_func_response(self.file, cog_serv_endpoint, cog_serv_key, max_pages)

    def get_pdfjson_images_text(self):
        max_pages = -1
        pdf_file_base64 = self.args.get('PDFFile')
        max_pages_str = self.args.get('MaxPages')
        if max_pages_str is not None:
            max_pages = int(max_pages_str)
        cog_serv_endpoint = self.env.get_secret_value('ImgCogServicesEndpoint')
        cog_serv_key = self.env.get_secret_value('ImgCogServicesKey')
        return create_pdf_func_response(base64.b64decode(pdf_file_base64),  cog_serv_endpoint, cog_serv_key, max_pages)

    def get_pdfblob_images_text(self):
        connect_str = self.env.get_parameter('AzStorageContainer')
        local_file_name = self.args.get('PDFFile')
        container_name = self.args.get('StorageSubDir')
        blob_service_client = BlobServiceClient.from_connection_string(connect_str)
        blob_client = blob_service_client.get_blob_client(container=container_name, blob=local_file_name)
        fstream = blob_client.download_blob()
        pdf_file = fstream.readall()

        max_pages_str = self.args.get('MaxPages')
        if max_pages_str is not None:
            max_pages = int(max_pages_str)
        else:
            max_pages = -1

        cog_serv_endpoint = self.env.get_secret_value('ImgCogServicesEndpoint')
        cog_serv_key = self.env.get_secret_value('ImgCogServicesKey')

        return create_pdf_func_response(pdf_file, cog_serv_endpoint, cog_serv_key, max_pages)






import os

WL_PROJECT_ROOT_ENV_VARIABLE = 'WL_PROJECT_ROOT'

def get_env_name():
    wl_project_root = os.environ.get(WL_PROJECT_ROOT_ENV_VARIABLE)
    if wl_project_root:
        env_name = 'DEV'
    else:
        env_name = 'TEST'
    return env_name


import io
from PIL import Image, ImageOps
import base64
from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
from wllib2.img_lib import img_utils

class ImgAzHandler():

    def __init__(self, app_env, input_args, input_file=None):
        self.env = app_env
        self.args = input_args
        self.file = input_file
        return 

    def get_image_text(self):
        cog_serv_endpoint = self.env.get_secret_value('ImgCogServicesEndpoint')
        cog_serv_key = self.env.get_secret_value('ImgCogServicesKey')
        az_img_ocr = img_utils.AzureImageOCR(endpoint=cog_serv_endpoint, key=cog_serv_key)
        img = Image.open(io.BytesIO(self.file))
        return az_img_ocr.get_image_text(img)






''' 
    WL_Az_http_func_handler is a class intended to be used by Azure Serverless Functions
    When invoked an Azure function will create an object of this class, pass
    to it the HttpRequest & HttpContext it has been passed.
    Then call the exec_request() method and return a composed result
'''  
import sys, traceback
import os
from importlib import reload
import pandas as pd
from datetime import date, datetime
import json
from wllib2.env_lib import WLAzAppEnv

def _proccess_request_inputs(http_req):
    ''' Takes an Azure function HttpRequest as an argument and extracts out all the parameters (get or post or json)
        and extracts the binary file if there is one.

    Args:
        http_req (Azure function HttpRequest): The HTTP request data

    Returns:
        req_params (Dictionary), req_file (binary file bytes): The data extracted from the request.
    '''
    req_params = {}
    req_file = None

    req_get_params = http_req.params
    for req_key in req_get_params:
        req_val = req_get_params.get(req_key)
        req_params[req_key] = req_val

    try:
        req_body = http_req.get_json()
    except ValueError:
        pass
    else:
        for req_key in req_body:
            req_val = req_body.get(req_key)
            req_params[req_key] = req_val

    try:
        file_bytes = http_req.get_body()
    except ValueError:
        pass
    else:
        req_file = file_bytes

    return req_params, req_file

def _dict_to_html(r, d, nametag="<strong>%s: </strong>", itemtag='<li>%s</li>',
             valuetag="%s", blocktag=('<ul>', '</ul>')):
    ''' Recursive function which takes a python dictionary type and converts it into a list of strings containing HTML which 
       , if joined together, will display the dictionary data on a web page.

    Args:
        r (list): the list of HTML strings
        d (dictionary): the dictionary source data
        nametag (str): html styling for nametags
        itemtag (str): html styling for itemtags
        valuetag (str): html styling for valuetags
        blocktag (str): html styling for blocktags
        
    Returns:
        nothing returned - the output is in the r list argument.
    '''   
    if isinstance(d, dict):
        r.append(blocktag[0])
        for k in d:
            v = d[k]
            name = nametag % k
            if isinstance(v, dict) or isinstance(v, list):
                r.append(itemtag % name)
                _dict_to_html(r, v)
            else:
                if isinstance(v, pd.DataFrame):
                    value = v.to_html()
                else:
                    value = valuetag % v
                r.append(itemtag % (name + value))
        r.append(blocktag[1])
    elif isinstance(d, list):
        r.append(blocktag[0])
        for i in d:
            if isinstance(i, dict) or isinstance(i, list):
                r.append(itemtag % " - ")
                _dict_to_html(r, i)
            else:
                r.append(itemtag % i)
    r.append(blocktag[1])
    return

def _json_serialize(obj):
    ''' JSON serializer for objects not serializable by default json code

    Args:
        obj (any python type): the thing to be serialised
        
    Returns:
        a serializable version of obj.
    '''
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial

    if isinstance(obj, date):
        serial = obj.isoformat()
        return serial

    if isinstance(obj, pd.DataFrame):
        serial = obj.to_dict()
        return serial

    if isinstance(obj, pd.Series):
        serial = obj.to_dict()
        return serial

    return obj.__dict__



class WL_Az_http_func_handler():

    def __init__(self, proj_root_path, http_request, function_context):
        ''' Instanciate a AZ Function handler object

        Args:
            proj_root_path (str): the root folder for the function where it is being run from
            http_request (azure function HttpRequest): the request data involved in the function call
            function_context (azure function HttpContext): the request context of the function call
            
        Returns:
            n/a
        '''
        func_name = function_context.function_name
        self.env = WLAzAppEnv.WLAzAppEnv(proj_root_path, func_name)

        self.funcdef_well_formed = False
        self.response_status = 500
        self.response_wrapper = {}

        if self.env:
            self.env.clear_app_log_messages()

            self.request_params, self.request_file = _proccess_request_inputs(http_request)

            if len(self.request_params):
                self.func_def = self.request_params.get('FuncDef')
                if self.func_def and len(self.func_def.split('.')) == 4:
                    f_parts =  self.func_def.split('.')
                    self.func_lib = f_parts[0]
                    self.func_classes = f_parts[1]
                    self.func_class = f_parts[2]
                    self.func_method = f_parts[3]
                    self.funcdef_well_formed = True

                if self.funcdef_well_formed: 
                    self.response_mode = self.request_params.get('ResponseMode')
                    self.response_wrapper['Handler'] = 'WL_Az_http_func_handler'    
                    self.response_wrapper['Function'] = self.func_def
                    self.response_wrapper['ResponseMode'] = self.response_mode

                    if self.request_file:
                        self.response_wrapper['FileUpload'] = True
                        self.response_wrapper['FileSize'] = len(self.request_file)
                    
                    self.exec_response = {}

                    self.request_method_args = {}
                    for key in self.request_params:
                        if key != 'FuncDef' and key != 'ResponseMode' and key != 'code':
                            self.request_method_args[key] = self.request_params.get(key)

                    readable_method_args = {}
                    for key in self.request_method_args:
                        val = self.request_method_args.get(key)
                        if not isinstance(val,bytes):
                            if sys.getsizeof(val) < 300:
                                readable_method_args[key] = val
                            else:
                                readable_method_args[key] = 'Long variables not included.'
                        else:
                            readable_method_args[key] = 'File binary contents not included.'

                    self.response_wrapper['FuncArgs'] = readable_method_args
                    self.env.log_message('FuncArgs: '+ json.dumps(readable_method_args))
                else:
                    mess = f'Function definition [{self.func_def}] was empty or malformed. Should be [Library_directory.Library_file.Class.Method].'
                    self.env.log_message('Error: '+ mess)
                    self.response_wrapper = {'Status':'Error','Message':mess}
                    self.response_mode = 'application/json'
                    self.response_status = 500
            else:
                mess = f'Failed to handle Function request - the request parameters are incorrectly formed.'
                self.env.log_message('Error: '+ mess)                
                self.response_wrapper = {'Status':'Error','Message':mess}
                self.response_mode = 'application/json'
                self.response_status = 500                
        else:
            mess = f'Failed to create instance of Function Handler Environment object.'
            self.env.log_message('Error: '+ mess)
            self.response_wrapper = {'Status':'Error','Message':mess}
            self.response_mode = 'application/json'
            self.response_status = 500
        return

    def exec_request(self):
        ''' Execute the Azure Function requested

        Args:
            n/a
            
        Returns:
            response_str (str): a json string containing the function response
        '''
        if self.funcdef_well_formed:
            str_func = f'from wllib2.{self.func_lib} import {self.func_classes}'
            str_func += f'\nreload({self.func_classes})'
            str_func += f'\nfunc_obj = {self.func_classes}.{self.func_class}(self.env,self.request_method_args,self.request_file)'
            str_func += f'\nself.exec_response = func_obj.{self.func_method}()'
            self.env.log_message(str_func)
            
            try:
                exec(str_func)
                self.response_status = 200
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                line_number = exc_traceback.tb_lineno
                traceback.clear_frames(exc_traceback)
                em = exc_type(exc_value).with_traceback(exc_traceback)

                str_err_mess = f'Function definition [{self.func_def}] could not be found or executed. '
                
                if self.env.get_env_name() != 'PROD':
                    str_err_mess += f'Error of type {exc_type} on line number {line_number} : message = [{em}]'
                    str_err_mess += f'Env log : {self.env.get_log_messages()}'

                self.env.log_message('Error: '+ str_err_mess)
                self.exec_response = {'Status':'Error','Message':str_err_mess}
                self.response_status = 500

            if self.response_mode == 'HTML':
                self.response_wrapper['ResponseData'] = self.exec_response
                html_response_list = []
                _dict_to_html(html_response_list, self.response_wrapper)
                response_str ='\n'.join(html_response_list)
            else:
                self.response_wrapper['ResponseData'] = self.exec_response
                response_str = json.dumps(self.response_wrapper, default=_json_serialize)
        else:
            response_str = json.dumps(self.response_wrapper)

        return response_str

    def response_mime_type(self):
        ''' Get the mime-type for the most recent call of exec_request()

        Args:
            n/a
            
        Returns:
            mime_type (str): the mime-type string for the most recent call of exec_request()
        '''
        if self.response_mode == 'HTML':
            mime_type = 'text/html'
        else:
            mime_type = 'application/json'
        return mime_type
    
    def get_response_status(self):
        ''' Get the status for the most recent call of exec_request()

        Args:
            n/a
            
        Returns:
            mime_type (int): the status number (200=ok, 500=fail) for the most recent call of exec_request()
        '''
        return self.response_status

    def get_request_params(self):
        ''' Get the request parameters (non-structural)

        Args:
            n/a
            
        Returns:
            request_params (dict): the request parameters passed to the function
        '''      
        return self.request_params

    def get_request_file(self):
        ''' Get the request file (if there is one)

        Args:
            n/a
            
        Returns:
            request_file (bytes): the request file passed to the function
        '''      
        return self.request_file

    def get_env(self):
        ''' Get the request environment object

        Args:
            n/a
            
        Returns:
            env (WLAzAppEnv(object)): the environment object created when this handler object was instanciated
        '''      
        return self.env


''' 
    Mock versions of Azure Function HttpRequest and HttpContext
    Intended to be used to test the WL_Az_http_func_handler Class
    Instances of these Classes can be created, populated and then passed 
    to the WL_Az_http_func_handler constructor
'''      
class MockHttpRequest():
    def __init__(self, req_params, req_file, req_json):
        if req_params:
            self.params = req_params
        else:
            self.params = {}

        if req_json:
            self.json = req_json
        else:
            self.json = {}
 
        self.bin_file = req_file
        return

    def get_json(self):
        return self.json
    
    def get_body(self):
        return self.bin_file

class MockHttpContext():
    def __init__(self, fname):
        self.function_name = fname


''' 
    WLAzAppEnv is a class intended to be used as an environment wrapper for Azure Serverless Function calls.
    It is created within the constructor of WL_Az_http_func_handler() class
    WLAzAppEnv establishes whether this execution is a Dev, TEST or PROD execution
    and delivers parameter value accordingly from two .ini config file (project and app)
    It offers a logging mechanism which write to Azure Application Insight for other code to use
    It accesses Azure KeyVault secret values if requested
    This class is designed as a singleton - meaning only one instance of this class may exist for an execution. 
    Any attempt to create a 2nd (or more) instance of this object will return the first instance created
'''
import sys
import os
import logging
from configparser import SafeConfigParser


WL_ENV_INI_FILENAME = 'wl_project_config.ini'
WL_APP_INI_FILENAME = 'wl_app_config.ini'
WL_ENV_PARAM = 'WL_APP_ENV'

def _get_parameter_value(env_p, app_p, env_name, param_name):
    try:
        wl_lib_path_env = env_p.get('ENV', param_name)
    except:
        wl_lib_path_env = None

    try:
        wl_lib_path_app = app_p.get(env_name, param_name)
    except:
        wl_lib_path_app = None

    return  wl_lib_path_app if wl_lib_path_app else wl_lib_path_env

def _init_logger():
    logger = logging.getLogger()
    h = logging.StreamHandler(sys.stdout)
    h.flush = sys.stdout.flush
    logger.addHandler(h)
    return logger

class WLAzAppEnv(object):
    _instance = None

    def __new__(self, proj_root_dir, app_name, using_azure_keyvault=True):
        ''' for first call creat and return an instance of WLAppEnv - for all future calls return the same object
            Access the config parser .ini files
            Initialise the log stash
            Connect to the Azure Key Vault

        Args: n/a
            
        Returns: the instance 
        '''          
        if self._instance is None:
            self._instance = super(WLAzAppEnv, self).__new__(self)

            self.wl_project_root = proj_root_dir
            self.app_name = app_name
            self.using_azure_keyvault = using_azure_keyvault

            proj_ini_file = self.wl_project_root + '/' + WL_ENV_INI_FILENAME
            self.env_parser = SafeConfigParser()
            self.env_parser.read(proj_ini_file)

            try:
                self.env_name = os.environ[WL_ENV_PARAM]
            except:
                self.env_name = self.env_parser.get('ENV', 'WLEnvironment')

            app_ini_file = proj_root_dir + '/' + app_name + '/' + WL_APP_INI_FILENAME
            self.app_parser = SafeConfigParser()
            self.app_parser.read(app_ini_file)

            self.env_log = f'Env = {self.env_name} Proj Root = {self.wl_project_root}; '
            self.env_log += f'\nproj_ini_file={proj_ini_file}; '
            self.env_log += f'\nApp ini = {app_ini_file}; '

            self.logger = _init_logger()
            self.logger.info(self.env_log)
            sys.stdout.flush

            self.app_log = ''
            self.log_mode = 'LOG_AND_PRINT'

            if self.using_azure_keyvault:
                from azure.identity import ClientSecretCredential
                from azure.keyvault.secrets import SecretClient
                tenant_id = self.env_parser.get('ENV', 'TenantID')
                client_id = self.env_parser.get('ENV', 'ClientID')
                client_secret = self.env_parser.get('ENV', 'ClientSecret')
                key_vault_name = self.env_parser.get('ENV', 'KeyVaultName')
                key_vault_uri = f'https://{key_vault_name}.vault.azure.net/'

                _credential = ClientSecretCredential(
                    tenant_id = tenant_id,
                    client_id = client_id,
                    client_secret = client_secret
                )

                try:
                    self.sc = SecretClient(vault_url=key_vault_uri, credential=_credential)
                except:
                    self.sc = None
                    self.env_log += f'Azure Key Vault [{key_vault_name}] could not be accessed; '
            else:
                self.sc = None

        return self._instance

    def get_env_name(self):
        ''' get the name of the environment one of DEV, TEST or PROD

        Args: n/a
            
        Returns:
            str: the name of the enviroment
        '''  
        return self.env_name

    def get_app_name(self):
        ''' get the name of the app

        Args: n/a
            
        Returns:
            str: the name of the app
        '''  
        return self.app_name

    def get_parameter(self, param_name):
        ''' get the vaue of a parameter - if directed get from Azure Key Vault

        Args: 
            param_name (str): the name of the parameter
            
        Returns:
            str: the value of the parameter
        '''  
        p_val = _get_parameter_value(self.env_parser, self.app_parser,  self.env_name, param_name)
        if p_val == '$azkv.secret' and self.using_azure_keyvault:
            p_val = self.get_secret_value(param_name)
        return p_val

    def log_message(self, mess, print_mode = 'PRINT'):
        ''' save a message in the log and print it if requested

        Args: 
            mess (str): log message
            print_mode (str): PRINT or NO_PRINT
            
        Returns:
            n/a
        '''  
        self.app_log += f'\n{mess};'
        if self.log_mode == 'LOG_AND_PRINT' and print_mode != 'NO_PRINT':
            self.logger.info(mess)
            sys.stdout.flush
        return

    def clear_app_log_messages(self):
        ''' delete the log message

        Args: 
            n/a
            
        Returns:
            n/a
        '''  
        self.app_log = ''
        return 

    def get_log_messages(self):
        ''' get all the log messages

        Args: 
            n/a
            
        Returns:
            str: all the log messages
        '''  
        return self.env_log + '\n' + self.app_log

    def get_secret_value(self, secret_name):
        ''' get a secret value from the Azure Key Vault

        Args: 
            secret_name (str): the name of the secret
            
        Returns:
            secret_val (str): value of the secret in the vault (None if not found)
        '''  
        secret_val = None
        if self.sc:
            try:
                secret_val = self.sc.get_secret(secret_name).value
            except:
                secret_val = None
        return secret_val

    



''' 
    WLAppEnv is a class intended to be used as an environment wrapper applications.
    It establishes whether this environment is a DEV, TEST or PROD execution
    and delivers parameter value accordingly from two .ini config files (env and app)
    It looks for a Windows environment valrible called WL_PROJECT_ROOT which tells it where the project root is - one up from this directory
    The wavelength_env.ini config file, which is in the same folder as this file is...
    - should be different in DEV/TEST/PROD setups
    - holds the wl_environment_name param name which defines whether the application is sunninga s DEV, TEST or PROD
    - holds parameters which are specific to the environment DEV/TEST/PROD
    The app_name argument to the constructor assume this is a the name of the sub-folder under the project _root where the appliction code is.
    It expects to file an ini file called app_name.ini in that folder where application specific parameter settings can be held.
    It offers a logging mechanism for other code to use
    It accesses Azure KeyVault secret values if requested
    This class is designed as a singleton - meaning only one instance of this class may exist for an execution. 
    Any attempt to create a 2nd (or more) instance of this object will return the first instance created
'''
import sys
import os
from configparser import SafeConfigParser
from wllib2.env_lib import env_utils

'''
WL_PROJECT_ROOT_ENV_VARIABLE = 'WL_PROJECT_ROOT'
WL_ENV_INI_FILE = 'wl_lib/env_utils/wavelength_env.ini'
WL_ENV_PARAM_SET = 'ENV'
WL_ENV_PARAM_NAME = 'wl_environment_name'
WL_LIB_ROOT_PARAM_NAME = 'wl_lib_root'
'''
WL_KEY_VAULT_PARAM_NAME = 'KeyVaultName'

def _get_parameter_value(app_p, env_name, param_name):
    ''' get a value of an environment parameter for the current DEV/TEST/PROD environment
        precendent is application parameters win over enviroment ones

    Args:
        env_p (SafeConfigParser): the enviroment config parser
        app_p (SafeConfigParser): the app config parser
        param_name (str): the requested parameter name
        
    Returns:
        str: the value of the parameter found, None if not found
    '''  
    '''
    try:
        wl_lib_path_env = env_p.get('ENV', param_name)
    except:
        wl_lib_path_env = None
    '''

    try:
        wl_lib_path_app = app_p.get(env_name, param_name)
    except:
        wl_lib_path_app = None

    return  wl_lib_path_app


class WLAppEnv(object):
    _instance = None

    def __new__(self, app_ini_file):
        ''' for first call creat and return an instance of WLAppEnv - for all future calls return the same object
            Access the config parser .ini files
            Initialise the log stash
            Connect to the Azure Key Vault

        Args: n/a
            
        Returns: the instance 
        '''  
        if self._instance is None:
            #print('Creating the object')
            self._instance = super(WLAppEnv, self).__new__(self)

            self.env_name = env_utils.get_env_name()

            '''
            self.wl_project_root = os.environ.get(WL_PROJECT_ROOT_ENV_VARIABLE)
            self.using_azure_keyvault = using_azure_keyvault

            self.env_parser = SafeConfigParser()
            wl_env_file_name = self.wl_project_root + '/' + WL_ENV_INI_FILE
            self.env_parser.read(wl_env_file_name)
            self.env_name = self.env_parser.get(WL_ENV_PARAM_SET, WL_ENV_PARAM_NAME)
 
            app_ini_file = self.wl_project_root + '/' + app_name + '/' + app_name + '.ini'
            '''

            self.app_parser = SafeConfigParser()
            self.app_parser.read(app_ini_file)

            '''
            wl_lib_root = _get_parameter_value(self.app_parser,  self.env_name, WL_LIB_ROOT_PARAM_NAME)
            sys.path.append(wl_lib_root)
            '''

            self.app_log = ''
            self.env_log = f'Env = {self.env_name}; '

            key_vault_name = _get_parameter_value(self.app_parser,  self.env_name, WL_KEY_VAULT_PARAM_NAME)
            if key_vault_name:
                self.using_azure_keyvault = True
            else:
                self.using_azure_keyvault = False

            if self.using_azure_keyvault:
                from azure.identity import ClientSecretCredential
                from azure.keyvault.secrets import SecretClient
                tenant_id = self.app_parser.get(self.env_name, 'TenantID')
                client_id = self.app_parser.get(self.env_name, 'ClientID')
                client_secret = self.app_parser.get(self.env_name, 'ClientSecret')
                key_vault_uri = f'https://{key_vault_name}.vault.azure.net/'

                _credential = ClientSecretCredential(
                    tenant_id = tenant_id,
                    client_id = client_id,
                    client_secret = client_secret
                )

                try:
                    self.sc = SecretClient(vault_url=key_vault_uri, credential=_credential)
                except:
                    self.sc = None
                    self.env_log += f'Azure Key Vault [{key_vault_name}] could not be accessed; '
            else:
                self.sc = None

            self.log_mode = 'LOG_AND_PRINT'
        return self._instance

    def get_env_name(self):
        ''' get the name of the environment one of DEV, TEST or PROD

        Args: n/a
            
        Returns:
            str: the name of the enviroment
        '''  
        return self.env_name

    def get_parameter(self, param_name):
        ''' get the vaue of a parameter - if directed get from Azure Key Vault

        Args: 
            param_name (str): the name of the parameter
            
        Returns:
            str: the value of the parameter
        '''  
        p_val = _get_parameter_value(self.app_parser,  self.env_name, param_name)
        if p_val == '$azkv.secret' and self.using_azure_keyvault:
            p_val = self.get_secret_value(param_name)
        return p_val

    def log_message(self, mess, print_mode = 'NO_PRINT'):
        ''' save a message in the log and print it if requested

        Args: 
            mess (str): log message
            print_mode (str): PRINT or NO_PRINT
            
        Returns:
            n/a
        '''  
        self.app_log += f'\n{mess};'
        if self.log_mode == 'LOG_AND_PRINT' and print_mode != 'NO_PRINT':
            print(mess)
        return

    def clear_app_log_messages(self):
        ''' delete the log message

        Args: 
            n/a
            
        Returns:
            n/a
        '''  
        self.app_log = ''
        return 

    def get_log_messages(self):
        ''' get all the log messages

        Args: 
            n/a
            
        Returns:
            str: all the log messages
        '''  
        return self.env_log + '\n' + self.app_log

    def get_secret_value(self, secret_name):
        ''' get a secret value from the Azure Key Vault

        Args: 
            secret_name (str): the name of the secret
            
        Returns:
            secret_val (str): value of the secret in the vault (None if not found)
        '''  
        secret_val = None
        if self.sc:
            try:
                secret_val = self.sc.get_secret(secret_name).value
            except:
                secret_val = None
        return secret_val


    



''' 
    db_utils.py is a set of utility functions and a class intended for SQL Server DB access
    The SQLDBConn class requires an instance of WLAppEnv or WLAzAppEnv to be past to its constructor
'''  
import pyodbc
import pandas as pd
from datetime import datetime
import calendar

def get_sql_formatted_datetime(dt):
    ''' Get a SQL acceptable datetime string given a python datetime type (uo to miliseconds)

    Args:
        dt (datetime): datetime to be converted
        
    Returns:
        sql datetime sting
    '''  
    return f'{dt.day}-{calendar.month_name[dt.month][0:3]}-{dt.year} {dt.hour}:{dt.minute}:{dt.second}.{round(dt.microsecond/1000.0)}'

def get_sql_formatted_date(dt):
    ''' Get a SQL acceptable date string given a python date 

    Args:
        dt (date): date to be converted
        
    Returns:
        sql date sting
    '''  
    return f'{dt.day}-{calendar.month_name[dt.month][0:3]}-{dt.year}'

def sql_strfield_arg(arg_obj):
    ''' Take a native python type and return a string representation of it which is 
        acceptable as an argument to a stored procedure or function in SQL 

    Args:
        arg_obj (native python type): argument object
        
    Returns:
        sql arg sting
    '''  
    if arg_obj is None:
        sql_arg_str = 'NULL'
    else:
        if isinstance(arg_obj, str):
            if str.strip(arg_obj) == '':
                sql_arg_str = 'NULL'
            else:
                sql_arg_str = '\'' + (str.strip(arg_obj)).replace('\'','\'\'') + '\''
        elif isinstance(arg_obj, int):
            sql_arg_str = str(arg_obj)
        elif isinstance(arg_obj, datetime):
            sql_arg_str = '\'' + get_sql_formatted_datetime(arg_obj) + '\''
        elif isinstance(arg_obj, float):
            sql_arg_str = str(arg_obj)
        else:
            sql_arg_str = str(arg_obj)
    return sql_arg_str

class SQLDBConn:

    def __init__(self, app_env, db_name=None):
        ''' Initialise SQL DB connection.
            Uses app_env to get connection string info.
            Success or otherwise of DB connection is written to the app_env log

        Args:
            n/a
            
        Returns:
            n/a
        '''  
        self.env = app_env
        self.server = app_env.get_parameter('DBServer')
        if db_name:
            self.database = db_name
        else:
            self.database = app_env.get_parameter('DataBase')
        self.driver =  app_env.get_parameter('DBDriver')
        self.authmode = app_env.get_parameter('DBAuthMode')
        self.is_connected = False

        if self.authmode == 'TRUSTED':
            db_conn_str = r'Driver='+self.driver+';Server='+self.server+';Database='+self.database+';Trusted_Connection=yes;'
        else:
            usr = app_env.get_parameter('DBusername')
            pwd = app_env.get_parameter('DBpwd')
            db_conn_str = r'DRIVER='+self.driver+';SERVER='+self.server+';PORT=1433;DATABASE='+self.database+';UID='+usr+';PWD='+ pwd

        self.env.log_message(f'Connecting with [{db_conn_str}]')
        try:
            self.db_connection = pyodbc.connect(db_conn_str)
            self.is_connected = True
        except pyodbc.Error as ex:
            sqlstate = ex.args[1]
            self.is_connected = False
        
        if self.is_connected:
            self.env.log_message('Connected to SQL Server successfully')
        else:
            self.env.log_message(f'Failed to conect to SQL Server - [{sqlstate}]')

        return

    def conn_ok(self):
        return self.is_connected

    def get_connection(self):
        return self.db_connection

    def execute_stored_proc(self, proc_name, arg_list):
        ''' Execute a SQL stored procedure and return the result
            Only intended for use with stored procedures which update data, 
            use fetch_sql_query() instead for stored procedures which simply fetch data
            Assumes that all stored procedures have a last arument as an OUTPUT argument 
            containing error messages and they retrun an int status value

        Args:
            proc_name (str): the name of the stored procedure
            arg_list (list): the list of the stored procedure arguments (must not include OUTPUT last argument)
            
        Returns:
            (return_status,return_error) (tuple): 
                return_status (int) : >= 0 for success <0 for failure
                return_error (str): '' for success and containing an error message if failed
        '''  
        if self.is_connected:
            sql_arg_list = []
            for arg in arg_list:
                sql_arg_list.append(sql_strfield_arg(arg))
            arg_str = ','.join(sql_arg_list)

            sql_proc_stmt = f'exec @ret = {proc_name} {arg_str},@err OUTPUT; '
            #print(f'sql_proc_stmt={sql_proc_stmt}')
            sql_proc_top = 'declare @ret int, @err varchar(255); '
            sql_tail_tail = 'select @ret as StatusReturn, @err as ErrorMessage;'

            sql_command = sql_proc_top + sql_proc_stmt + sql_tail_tail
            
            cursor = self.db_connection.cursor()
            cursor.execute(sql_command)
            while 1:
                row = cursor.fetchone()
                if row is not None:
                    return_status = row[0]
                    return_error = row[1]                
                else:               
                    break

            self.db_connection.commit()
        else:
            return_status = -1
            return_error = 'DB not connected.'

        return (return_status,return_error)

    def fetch_sql_query(self,sql_query):
        ''' Execute a SQL query which fetches data and return the result as a dataframe
            The SQL query must only return a single dataset result

        Args:
            sql_query (str): the SQL query string
            
        Returns:
            ret_df (dataframe): the query result data set
        '''  
        if self.is_connected:
            cursor = self.db_connection.cursor()
            cursor.execute(sql_query)
            dataset = cursor.fetchall()
            dataset_info = cursor.description
            query_cols = []
            for col in dataset_info:
                query_cols.append(col[0])
            ret_df = pd.DataFrame.from_records(dataset, columns = query_cols)
        else:
            ret_df = None
        return ret_df

    def disconnect(self):
        ''' Disconnect from the database. Use when finished querying otherwise connection will remain.
            Write outcome to app_env log
        '''  
        if self.is_connected:
            self.env.log_message('Disconnecting from SQL Server')
            self.db_connection.close()
            self.is_connected = False
        else:
            self.env.log_message('Cannot disconnect from DB as not currently connected.')
        return

    def __del__(self):
        ''' Code added to class destructor to make sure that instances of this object
            disconnect from the DB when they go out of scope
        '''  
        if self.is_connected:
            self.env.log_message('Disconnecting from SQL Server')
            self.db_connection.close()
            self.is_connected = False
        return


from setuptools import setup, find_packages

setup(
   name='wavelengthlib2',
   version='0.1.21',
   description='Wavelength library 2',
   author='Simon',
   author_email='simon.elliott@wavelength.law',
   classifiers=[
            'Intended Audience :: Developers',
            'Programming Language :: Python'
          ],
   install_requires=[
         'azure-functions',
         'azure-storage-blob',
         'azure-cognitiveservices-vision-computervision==0.5.0',
         'azure-common==1.1.25',
         'azure-core==1.6.0',
         'azure-identity==1.3.1',
         'azure-keyvault-certificates==4.1.0',
         'azure-keyvault-keys==4.1.0',
         'azure-keyvault-secrets==4.1.0',
         'pyodbc==4.0.30',
         'python-dateutil',
         'configparser',
         'pandas',
         'numpy',
         'datetime',
         'requests',
         'Unidecode',
         'openpyxl==3.0.3',
         'python_docx==0.8.10',
         'PyPDF2==1.26.0',
         'pdf2image==1.6.0',
         'matplotlib',
         'docx',
         'PyMuPDF==1.16.18',
         'python_dateutil',
         'Pillow'
   ],
   python_requires='>=3.7',
   package_dir={'': '.'},
   packages=['wllib2','wllib2.lib_tests','wllib2.env_lib','wllib2.az_lib','wllib2.db_lib','wllib2.pdf_lib','wllib2.img_lib','wllib2.test_lib']
)


import os
import io
import sys
import struct
from PIL import Image, ImageOps
import sys, time, requests, traceback
import PyPDF2
from PyPDF2 import PdfFileReader
import json

'''
import os
import io
import sys
import struct
from PIL import Image, ImageOps
import sys, time, requests, traceback
import PyPDF2
from PyPDF2 import PdfFileReader
import base64
from collections import namedtuple
import warnings
import json

from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
from azure.cognitiveservices.vision.computervision import ComputerVisionClient
from azure.cognitiveservices.vision.computervision.models import TextOperationStatusCodes
from msrest.authentication import CognitiveServicesCredentials

warnings.filterwarnings("ignore")

PdfImage = namedtuple('PdfImage', ['data', 'format','image_name'])


'''
from azure.cognitiveservices.vision.computervision import ComputerVisionClient
from azure.cognitiveservices.vision.computervision.models import TextOperationStatusCodes
from msrest.authentication import CognitiveServicesCredentials

'''
def convert_gray(img, preprocessing):
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    if preprocessing == 'thresh':
        gray = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]
    elif preprocessing == 'blur':
        gray = cv2.medianBlur(gray, 3)
    return gray
'''

def tiff_header_for_CCITT(width, height, img_size, CCITT_group=4):
    # http://www.fileformat.info/format/tiff/corion.htm
    fields = 8
    tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H' + 'HHLL' * fields + 'L'
    return struct.pack(tiff_header_struct,
                       b'II',  # Byte order indication: Little indian
                       42,  # Version number (always 42)
                       8,  # Offset to first IFD
                       fields,  # Number of tags in IFD
                       256, 4, 1, width,  # ImageWidth, LONG, 1, width
                       257, 4, 1, height,  # ImageLength, LONG, 1, lenght
                       258, 3, 1, 1,  # BitsPerSample, SHORT, 1, 1
                       259, 3, 1, CCITT_group,  # Compression, SHORT, 1, 4 = CCITT Group 4 fax encoding
                       262, 3, 1, 0,  # Threshholding, SHORT, 1, 0 = WhiteIsZero
                       # StripOffsets, LONG, 1, len of header
                       273, 4, 1, struct.calcsize(tiff_header_struct),
                       278, 4, 1, height,  # RowsPerStrip, LONG, 1, length
                       279, 4, 1, img_size,  # StripByteCounts, LONG, 1, size of image
                       0  # last IFD
                       )


def image_to_byte_array(img:Image, img_format):
    imgByteArrIOStream = io.BytesIO()
    img.save(imgByteArrIOStream, format=img_format)
    imgBytes = imgByteArrIOStream.getvalue()
    return imgBytes


class AzureImageOCR():
    def __init__(self, endpoint=None, key=None):
        if endpoint is not None:
            self.endpoint = endpoint
        if key is not None:
            self.key = key
        self.client = ComputerVisionClient(self.endpoint, CognitiveServicesCredentials(self.key))
        return
        
    def get_image_text(self, img, preprocess='blur'):
        '''Performing OCR using Azure Cognitive API. 
        See https://docs.microsoft.com/en-gb/azure/cognitive-services/computer-vision/quickstarts/python-hand-text?tabs=version-3
        
        :params image: image binary 
        :returns extracted_text: string of the recognized text
        '''
        image_bytes = image_to_byte_array(img, 'JPEG')
        text_recognition_url = self.endpoint +  "/vision/v3.0-preview/read/analyze"
        headers = {'Ocp-Apim-Subscription-Key': self.key, 'Content-Type': 'application/octet-stream'}
        params = {'language': 'en'}
        response = requests.post(text_recognition_url, headers=headers, params=params, data=image_bytes)
        response.raise_for_status()
        
        # Hold the URL used to retreive the recognized text
        operation_url = response.headers['Operation-Location']

        # Poll to wait for completion
        analysis = {}
        poll = True
        while (poll):
            response_final = requests.get(operation_url, headers=headers)
            analysis = response_final.json()
            time.sleep(1)
            if ('analyzeResult' in analysis):
                poll = False
            if ('status' in analysis and analysis['status'] == 'failed'):
                poll = False
        
        # Extract the recognized text
        extracted_text = ''
        if ('analyzeResult' in analysis):
            text = [x['text'] for x in analysis['analyzeResult']['readResults'][0]['lines']]
            extracted_text = ' '.join(x for x in text)

        # Extract the word bounding box and text
        return extracted_text




#import os
import io
#import sys
#import struct
from PIL import Image, ImageOps
#import sys, time, requests, traceback
import PyPDF2
from PyPDF2 import PdfFileReader
#import base64
#from collections import namedtuple
import warnings
#import json

from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
from wllib2.img_lib import img_utils

#PdfImage = namedtuple('PdfImage', ['data', 'format','image_name'])

warnings.filterwarnings("ignore")

img_modes = {'/DeviceRGB': 'RGB', '/DefaultRGB': 'RGB',
             '/DeviceCMYK': 'CMYK', '/DefaultCMYK': 'CMYK',
             '/DeviceGray': 'L', '/DefaultGray': 'L',
             '/Indexed': 'P'}


def extract_image_objects_from_pdf_page(xObject):
    image_obj_list = []
    xobj_res_xobj_obj = None
    
    if xObject:
        xobj_res = xObject.get('/Resources')
        if xobj_res:
            xobj_res_xobj = xobj_res.get('/XObject')
            if xobj_res_xobj:
                xobj_res_xobj_obj = xobj_res_xobj.getObject()

    #xObject = xObject['/Resources']['/XObject'].getObject()

    if xobj_res_xobj_obj:
        for obj in xobj_res_xobj_obj:
            # xObj_this_obj = xObject.get(obj)
            xObj_this_obj = xobj_res_xobj_obj[obj]

            subtype_is_image = False
            if xObj_this_obj:
                if xObj_this_obj.get('/Subtype') == '/Image':
                    subtype_is_image = True

            if subtype_is_image:
                size = (xObj_this_obj.get('/Width'), xObj_this_obj.get('/Height'))
                data = xObj_this_obj._data
                xobj_obj = xObj_this_obj
                color_space = xobj_obj.get('/ColorSpace')
                if '/FlateDecode' in xobj_obj.get('/Filter'):
                    if isinstance(color_space, PyPDF2.generic.ArrayObject) and color_space[0] == '/Indexed':
                        color_space, _, _, _ = [v.getObject() for v in color_space] 
                    mode = img_modes[color_space]
                    data = xObj_this_obj.getData() 
                    img = Image.frombytes(mode, size, data)
                    if color_space == '/Indexed':
                        img.putpalette(lookup.getData())
                        img = img.convert('RGB')
                    imgByteArr = io.BytesIO()
                    img.save(imgByteArr,format='PNG')
                    if img.width > img.height:
                        img = img.rotate(90, expand=True)
                    image_obj_list.append(img)                
                elif '/DCTDecode' in xobj_obj.get('/Filter') or '/JPXDecode' in xobj_obj.get('/Filter'):
                    img = Image.open(io.BytesIO(data))
                    if img.width > img.height:
                        img = img.rotate(90, expand=True)
                    image_obj_list.append(img)
                elif '/CCITTFaxDecode' in xobj_obj.get('/Filter'):
                    decode_params = xObj_this_obj.get('/DecodeParms')
                    if decode_params.get('/K') == -1:
                        CCITT_group = 4
                    else:
                        CCITT_group = 3
                    data = xObj_this_obj._data 
                    img_size = len(data)
                    tiff_header = img_utils.tiff_header_for_CCITT(size[0], size[1], img_size, CCITT_group)
                    img = Image.open(io.BytesIO(tiff_header + data))

                    if img.width > img.height:
                        img = img.rotate(90, expand=True)
                    
                    image_obj_list.append(img)
            else:
                image_obj_list += extract_image_objects_from_pdf_page(xObj_this_obj)
    
    return image_obj_list

def get_pdf_images_from_pypdf2(py_pdf2_file, max_pages):
    pdf_file_image_list = []
    for p in range(max_pages):
        page_image_list = extract_image_objects_from_pdf_page(py_pdf2_file.getPage(p))
        for img in page_image_list:
            pdf_file_image_list.append(img)
    return pdf_file_image_list


def get_text_from_pdfbytes_images(pdf_bytes, cog_serv_endpoint, cog_serv_key, max_pages = -1):
    wholetext = ''
    number_of_pages_in_pdf = 0

    py_pdf2_file = PyPDF2.PdfFileReader(io.BytesIO(pdf_bytes))
    number_of_pages_in_pdf = py_pdf2_file.getNumPages()

    if max_pages == -1:
        max_pages = number_of_pages_in_pdf
    else:
        max_pages if max_pages > number_of_pages_in_pdf else number_of_pages_in_pdf
    
    pdf_page_images = get_pdf_images_from_pypdf2(py_pdf2_file, max_pages)

    if len(pdf_page_images):
        az_ocr = img_utils.AzureImageOCR(endpoint=cog_serv_endpoint, key=cog_serv_key)
        for page_img in pdf_page_images:
            wholetext += az_ocr.get_image_text(page_img)

    return wholetext, len(pdf_page_images), number_of_pages_in_pdf



